/***************************************************************************/
/*Name:          SCENT: Preprocessor                                       */
/*Location:      Undefined                                                 */
/*Author:        Justin A. Strauss                                         */
/*Organization:  Kaiser Permanente Southern California                     */
/*Purpose:       Prepare text for analysis                                 */
/*Modified:                                                                */
/***************************************************************************/

%include "&scentPreproc\Source\preprocessText_Tokenize.sas";

/*
MACRO PARAMETER NOTES:
data = name/library of source dataset containing report text
var = name of variable within source dataset containing report text
patterns = path to spreadsheet containing regular expression patterns
out = optional name/library of output dataset, if different from source
*/

%macro preprocessText(data=,out=,var=,varout="scent_preprocessed",patterns=,debug="no",obs=) / store secure;

/*remove quotation marks surrounding supplied parameters*/
%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
%else %return;
%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
%else %let out=&data;
%if (%length(&var)>0) %then %let var=%sysfunc(dequote(&var));
%else %return;
%if (%length(&varout)>0) %then %let varout=%sysfunc(dequote(&varout));
%if (%length(&patterns)>0) %then %let patterns=%sysfunc(dequote(&patterns));
%else %return;
%if (%length(&debug)>0) %then %let debug=%sysfunc(dequote(&debug));
%if (%length(&obs)>0) %then %let obs=%sysfunc(dequote(&obs));

/*read in the regular expression patterns used in preprocessing*/
libname xls "&patterns";
    data patterns;
        set xls.'SAS$'n(rename=(pattern=pattern_text enabled=pattern_enabled));
		length pattern 8;
		call missing(pattern);
    run;
libname xls clear;

/*limit dataset records, if requested*/
%if (%length(&obs)>0) %then %let obs=%str(%(obs=)&obs%str());
data &out(compress=yes);
	set &data &obs;
run;

/*compress option is important due to large text variables*/
data &out(compress=yes drop=log_: pattern: prevtext _:)
	%if (%upcase(&debug)~=NO) %then %do;
		pattern_matches(compress=yes keep=link_value accessiondate txt_final &varout pattern_match:)
		pattern_matchcounts(keep=_pattern_match:)
	%end;
;

    /*read in source data and store record count*/
    set &out nobs=log_count;
    /*create/retain variables for completion progress updates*/
    retain log_lastpct log_start;
    /*new variable to store &varout text*/
    length &varout prevtext $20000;

	if (_N_=1) then do;
		log_start=datetime();
		log_lastpct=0;
		array _pattern_match {100} 8;
			array pattern_match {100} 8;
		/*read regular expression patterns into hash table*/
		length pattern_id pattern pattern_enabled 8 pattern_text $500;
		drop pattern_id pattern pattern_enabled pattern_text;
		declare hash patterns(dataset:"patterns", ordered:"a");
		declare hiter hlookup('patterns');
		call patterns.defineKey('pattern_id');
		call patterns.defineData('pattern_id');
		call patterns.defineData('pattern_enabled');
		call patterns.defineData('pattern_text');
		call patterns.defineData('pattern');
		call patterns.defineDone();
		call missing(pattern_id, pattern, pattern_enabled, pattern_text);
		_rc=hlookup.first();
		do while (_rc=0 | _remove);
			if (_remove) then do;
				call patterns.remove(key: _remove_id);
				if (_rc~=0) then leave;
			end;
			_remove=ifn(pattern_enabled,0,1,0);
			_remove_id=ifn(pattern_enabled,.,pattern_id,.);
			if (pattern_enabled) then call patterns.replace(key: pattern_id, data: pattern_id, data: pattern_enabled, data: pattern_text, data: prxparse("S" || strip(pattern_text)));
			_rc=hlookup.next();
		end;
	end;

    /*assign original report text to new variable*/
    &varout=&var;
    prevtext=&varout;

    _rc=hlookup.first();
    do while (_rc=0);
        &varout=prxchange(pattern, -1, strip(&varout));
        if (strip(upcase(&varout))~=strip(upcase(prevtext))) then do;
           _pattern_match{pattern_id}+1;
		   pattern_match{pattern_id}=1;
            prevtext=&varout;
        end;
		_rc=hlookup.next();
    end;

    /*calculate % of overall records processed thus far*/
    log_pctdone=(_N_/log_count)*100;
    /*provide status update in log if >=5% more records processed since previous message*/
    if (log_pctdone>=(log_lastpct+5) & log_pctdone<100) then do;
        log_lastpct=round((_N_/log_count)*100,1.0);
        log_timepass=datetime()-log_start;
        log_n=put(_N_,8.0);
        /*determine whether to provide lapsed time in seconds or minutes*/
        if (log_timepass<=60) then log_text=strip(put(_N_,COMMA8.0)) || " of " || strip(put(log_count,COMMA8.0)) || " (" || strip(put(log_lastpct,8.0)) || "%)" || " in " || strip(put(log_timepass,4.1)) || " seconds.";
        else log_text=strip(put(_N_,COMMA8.0)) || " of " || strip(put(log_count,COMMA8.0)) || " (" || strip(put(log_lastpct,8.0)) || "%)" || " in " || strip(put(log_timepass/60,5.1)) || " minutes.";
        put "NOTE: Finished coding observation #" log_text;
    end;

    output &out;
	%if (%upcase(&debug)~=NO) %then %do;
		output pattern_matches;
    	if (_N_=log_count) then output pattern_matchcounts;
	%end;
run;

/*output pattern match diagnostics if debugging*/
%if (%upcase(&debug)~=NO) %then %do;
	proc transpose data=pattern_matchcounts out=pattern_matchcounts(rename=(_name_=name col1=count)); run;
	data pattern_matchcounts;
	    length pattern_id 8;
	    set pattern_matchcounts;
	    pattern_id=_N_;
    	if (missing(count)) then count=0;
	run;
	proc sql;
	    create table pattern_stats as
        	select a.*, b.count
        	from patterns as a left join pattern_matchcounts as b
        	on (a.pattern_id=b.pattern_id);
    	drop table patterns, pattern_matchcounts;
	quit;
%end; %else %do;
	proc sql;
		alter table pattern_stats drop pattern;
    	drop table patterns;
	quit;
%end;

%mend preprocessText;
